import json
import logging
import os
import re
from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

SimCSEExample = namedtuple(
    "SimCSEExample",
    [
        "sequence",
        "input_ids",
        "segment_ids",
        "attention_mask",
        "pos_sequence",
        "pos_input_ids",
        "pos_segment_ids",
        "pos_attention_mask",
        "neg_sequence",
        "neg_input_ids",
        "neg_segment_ids",
        "neg_attention_mask",
    ],
)


class SimCSEDataset:
    """Dataset builder for SimCSE models."""

    @classmethod
    def from_tfrecord_files(
        cls,
        input_files,
        batch_size=64,
        repeat=None,
        with_pos_sequence=False,
        with_neg_sequence=False,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=True,
        **kwargs
    ):
        dataset = cls._read_tfrecord(
            input_files, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        return cls._build(
            dataset,
            batch_size=batch_size,
            repeat=repeat,
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            max_sequence_length=max_sequence_length,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            buffer_size=buffer_size,
            seed=seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
            pad_id=pad_id,
            auto_shard_policy=auto_shard_policy,
            drop_remainder=drop_remainder,
            **kwargs,
        )

    @classmethod
    def from_jsonl_files(cls, input_file, vocab_file, **kwargs):
        examples = cls.jsonl_to_examples(input_file, vocab_file, **kwargs)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls,
        examples: List[SimCSEExample],
        with_pos_sequence=False,
        with_neg_sequence=False,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=True,
        verbose=True,
        **kwargs
    ):
        logging.info("Number of examples: %d", len(examples))
        cls._show_examples(examples, n=5, verbose=verbose, **kwargs)
        dataset = cls._zip_dataset(
            examples, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        return cls._build(
            dataset,
            batch_size=batch_size,
            repeat=repeat,
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            max_sequence_length=max_sequence_length,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            buffer_size=buffer_size,
            seed=seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
            pad_id=pad_id,
            auto_shard_policy=auto_shard_policy,
            drop_remainder=drop_remainder,
            **kwargs,
        )

    @classmethod
    def _build(
        cls,
        dataset,
        with_pos_sequence=False,
        with_neg_sequence=False,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=True,
        **kwargs
    ):
        dataset = cls._filter_dataset(
            dataset,
            max_sequence_length=max_sequence_length,
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            **kwargs,
        )
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(
            buffer_size=buffer_size,
            seed=seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
        )
        dataset = cls._bucketing(
            dataset,
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(
            dataset, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, verbose=True, **kwargs):
        if not verbose:
            return
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("No.%d example: %s", i, examples[i])

    @classmethod
    def _zip_dataset(cls, examples: List[SimCSEExample], with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        """zip dataset"""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        if with_neg_sequence:
            dataset = tf.data.Dataset.zip(
                (
                    _to_dataset([e.input_ids for e in examples]),
                    _to_dataset([e.segment_ids for e in examples]),
                    _to_dataset([e.attention_mask for e in examples]),
                    _to_dataset([e.pos_input_ids for e in examples]),
                    _to_dataset([e.pos_segment_ids for e in examples]),
                    _to_dataset([e.pos_attention_mask for e in examples]),
                    _to_dataset([e.neg_input_ids for e in examples]),
                    _to_dataset([e.neg_segment_ids for e in examples]),
                    _to_dataset([e.neg_attention_mask for e in examples]),
                )
            )
        elif with_pos_sequence:
            dataset = tf.data.Dataset.zip(
                (
                    _to_dataset([e.input_ids for e in examples]),
                    _to_dataset([e.segment_ids for e in examples]),
                    _to_dataset([e.attention_mask for e in examples]),
                    _to_dataset([e.pos_input_ids for e in examples]),
                    _to_dataset([e.pos_segment_ids for e in examples]),
                    _to_dataset([e.pos_attention_mask for e in examples]),
                )
            )
        else:
            dataset = tf.data.Dataset.zip(
                (
                    _to_dataset([e.input_ids for e in examples]),
                    _to_dataset([e.segment_ids for e in examples]),
                    _to_dataset([e.attention_mask for e in examples]),
                )
            )
        return dataset

    @classmethod
    def _filter_dataset(
        cls, dataset, max_sequence_length=512, with_pos_sequence=False, with_neg_sequence=False, **kwargs
    ):
        if with_neg_sequence:
            dataset = dataset.filter(
                lambda a, b, c, e, f, g, h, i, j: tf.logical_and(
                    tf.size(a) <= max_sequence_length,
                    tf.logical_and(tf.size(e) <= max_sequence_length, tf.size(h) <= max_sequence_length),
                )
            )
        elif with_pos_sequence:
            dataset = dataset.filter(
                lambda a, b, c, e, f, g: tf.logical_and(
                    tf.size(a) <= max_sequence_length, tf.size(e) <= max_sequence_length
                )
            )
        else:
            dataset = dataset.filter(lambda a, b, c: tf.size(a) <= max_sequence_length)
        return dataset

    @classmethod
    def _bucketing(
        cls,
        dataset,
        with_pos_sequence=False,
        with_neg_sequence=False,
        batch_size=64,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=True,
        **kwargs
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_sizes) should equals len(bucket_doundaries) + 1"

        pad = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        if with_neg_sequence:
            dataset = dataset.apply(
                tf.data.experimental.bucket_by_sequence_length(
                    element_length_func=lambda a, b, c, e, f, g, h, i, j: tf.maximum(
                        tf.size(a), tf.maximum(tf.size(e), tf.size(h))
                    ),
                    bucket_boundaries=bucket_boundaries,
                    bucket_batch_sizes=bucket_batch_sizes,
                    padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,]),
                    padding_values=(pad, pad, pad, pad, pad, pad, pad, pad, pad),
                    drop_remainder=drop_remainder,
                )
            ).prefetch(tf.data.AUTOTUNE)
        elif with_pos_sequence:
            dataset = dataset.apply(
                tf.data.experimental.bucket_by_sequence_length(
                    element_length_func=lambda a, b, c, e, f, g: tf.maximum(tf.size(a), tf.size(e)),
                    bucket_boundaries=bucket_boundaries,
                    bucket_batch_sizes=bucket_batch_sizes,
                    padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,]),
                    padding_values=(pad, pad, pad, pad, pad, pad),
                    drop_remainder=drop_remainder,
                )
            ).prefetch(tf.data.AUTOTUNE)
        else:
            dataset = dataset.apply(
                tf.data.experimental.bucket_by_sequence_length(
                    element_length_func=lambda a, b, c: tf.size(a),
                    bucket_boundaries=bucket_boundaries,
                    bucket_batch_sizes=bucket_batch_sizes,
                    padded_shapes=([None, ], [None, ], [None, ]),
                    padding_values=(pad, pad, pad),
                    drop_remainder=drop_remainder,
                )
            ).prefetch(tf.data.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if with_neg_sequence:
            dataset = dataset.map(
                lambda a, b, c, e, f, g, h, i, j: (
                    {
                        "input_ids": a,
                        "segment_ids": b,
                        "attention_mask": c,
                        "pos_input_ids": e,
                        "pos_segment_ids": f,
                        "pos_attention_mask": g,
                        "neg_input_ids": h,
                        "neg_segment_ids": i,
                        "neg_attention_mask": j,
                    },
                    None,
                ),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
            return dataset
        elif with_pos_sequence:
            dataset = dataset.map(
                lambda a, b, c, e, f, g: (
                    {
                        "input_ids": a,
                        "segment_ids": b,
                        "attention_mask": c,
                        "pos_input_ids": e,
                        "pos_segment_ids": f,
                        "pos_attention_mask": g,
                    },
                    None,
                ),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
        else:
            dataset = dataset.map(
                lambda x, y, z: ({"input_ids": x, "segment_ids": y, "attention_mask": z}, None),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None, **kwargs):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _read_tfrecord(cls, input_files, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        if len(input_files) == 1:
            dataset = tf.data.TFRecordDataset(input_files)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(input_files)
            dataset = dataset.interleave(
                lambda x: tf.data.TFRecordDataset(x),
                cycle_length=len(input_files),
                num_parallel_calls=tf.data.AUTOTUNE,
            )
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        if with_neg_sequence:
            features.update(
                {
                    "neg_input_ids": tf.io.VarLenFeature(tf.int64),
                    "neg_segment_ids": tf.io.VarLenFeature(tf.int64),
                    "neg_attention_mask": tf.io.VarLenFeature(tf.int64),
                    "pos_input_ids": tf.io.VarLenFeature(tf.int64),
                    "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
                    "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
                }
            )
        elif with_pos_sequence:
            features.update(
                {
                    "pos_input_ids": tf.io.VarLenFeature(tf.int64),
                    "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
                    "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
                }
            )
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)

        if with_neg_sequence:
            dataset = dataset.map(
                lambda x: (
                    tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["neg_input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["neg_segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["neg_attention_mask"]), tf.int32),
                ),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
        elif with_pos_sequence:
            dataset = dataset.map(
                lambda x: (
                    tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
                ),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
        else:
            dataset = dataset.map(
                lambda x: (
                    tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                    tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                ),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def jsonl_to_tfrecord(
        cls, input_files, vocab_file, output_files, with_pos_sequence=False, with_neg_sequence=False, **kwargs
    ):
        examples = cls.jsonl_to_examples(
            input_files, vocab_file, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        cls.examples_to_tfrecord(
            examples, output_files, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        logging.info("Convert data from jsonl to tfrecord finished.")

    @classmethod
    def jsonl_to_examples(cls, input_files, vocab_file, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        examples = []
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.get("do_lower_case", True))
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = json.loads(line)
                    example = cls._instance_to_example(
                        instance,
                        tokenizer,
                        with_pos_sequence=with_pos_sequence,
                        with_neg_sequence=with_neg_sequence,
                        **kwargs,
                    )
                    examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def _instance_to_example(
        cls, instance, tokenizer: BertWordPieceTokenizer, with_pos_sequence=False, with_neg_sequence=False, **kwargs
    ):
        sequence = instance[kwargs.get("sequence_key", "sequence")]
        seq_encoding = tokenizer.encode(sequence)
        if with_neg_sequence:
            pos_sequence = instance[kwargs.get("pos_sequence_key", "pos_sequence")]
            pos_encoding = tokenizer.encode(pos_sequence)
            neg_sequence = instance[kwargs.get("neg_sequence_key", "neg_sequence")]
            neg_encoding = tokenizer.encode(neg_sequence)
            example = SimCSEExample(
                sequence=sequence,
                input_ids=seq_encoding.ids,
                segment_ids=seq_encoding.type_ids,
                attention_mask=seq_encoding.attention_mask,
                pos_sequence=pos_sequence,
                pos_input_ids=pos_encoding.ids,
                pos_segment_ids=pos_encoding.type_ids,
                pos_attention_mask=pos_encoding.attention_mask,
                neg_sequence=neg_sequence,
                neg_input_ids=neg_encoding.ids,
                neg_segment_ids=neg_encoding.type_ids,
                neg_attention_mask=neg_encoding.attention_mask,
            )
            return example
        if with_pos_sequence:
            pos_sequence = instance[kwargs.get("pos_sequence_key", "pos_sequence")]
            pos_encoding = tokenizer.encode(pos_sequence)
            example = SimCSEExample(
                sequence=sequence,
                input_ids=seq_encoding.ids,
                segment_ids=seq_encoding.type_ids,
                attention_mask=seq_encoding.attention_mask,
                pos_sequence=pos_sequence,
                pos_input_ids=pos_encoding.ids,
                pos_segment_ids=pos_encoding.type_ids,
                pos_attention_mask=pos_encoding.attention_mask,
                neg_sequence=None,
                neg_input_ids=None,
                neg_segment_ids=None,
                neg_attention_mask=None,
            )
            return example
        example = SimCSEExample(
            sequence=sequence,
            input_ids=seq_encoding.ids,
            segment_ids=seq_encoding.type_ids,
            attention_mask=seq_encoding.attention_mask,
            pos_sequence=None,
            pos_input_ids=None,
            pos_segment_ids=None,
            pos_attention_mask=None,
            neg_sequence=None,
            neg_input_ids=None,
            neg_segment_ids=None,
            neg_attention_mask=None,
        )
        return example

    @classmethod
    def examples_to_tfrecord(cls, examples, output_files, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if isinstance(output_files, str):
            output_files = [output_files]
        writers = [tf.io.TFRecordWriter(f) for f in output_files]
        idx = 0
        for example in examples:
            tfrecord_example = cls._example_to_tfrecord(
                example, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
            )
            writers[idx].write(tfrecord_example.SerializeToString())
            idx += 1
            idx = idx % len(writers)
        for w in writers:
            w.close()

    @classmethod
    def _example_to_tfrecord(cls, example: SimCSEExample, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
        }
        if with_neg_sequence:
            feature.update(
                {
                    "pos_input_ids": cls._int64_feature([int(x) for x in example.pos_input_ids]),
                    "pos_segment_ids": cls._int64_feature([int(x) for x in example.pos_segment_ids]),
                    "pos_attention_mask": cls._int64_feature([int(x) for x in example.pos_attention_mask]),
                    "neg_input_ids": cls._int64_feature([int(x) for x in example.neg_input_ids]),
                    "neg_segment_ids": cls._int64_feature([int(x) for x in example.neg_segment_ids]),
                    "neg_attention_mask": cls._int64_feature([int(x) for x in example.neg_attention_mask]),
                }
            )
        elif with_pos_sequence:
            feature.update(
                {
                    "pos_input_ids": cls._int64_feature([int(x) for x in example.pos_input_ids]),
                    "pos_segment_ids": cls._int64_feature([int(x) for x in example.pos_segment_ids]),
                    "pos_attention_mask": cls._int64_feature([int(x) for x in example.pos_attention_mask]),
                }
            )
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _int64_feature(cls, values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=values))



import os
import unittest

from transformers_keras.token_classification.callback import SeqEvalForTokenClassification
from transformers_keras.token_classification.models import BertForTokenClassification

BERT_PATH = os.path.join(os.environ["CHINESE_BERT_PATH"], "chinese_roberta_wwm_ext_L-12_H-768_A-12")
VOCAB_PATH = os.path.join(BERT_PATH, "vocab.txt")


class CallbackTest(unittest.TestCase):
    """Callback test"""

    def test_seqeval_for_token_classification(self):
        callback = SeqEvalForTokenClassification.from_conll_files(
            "testdata/conll.txt", "testdata/vocab.bert.txt", "testdata/labels.txt", sep="\\s+"
        )
        model = BertForTokenClassification.from_pretrained(BERT_PATH, override_params={"num_labels": 3})
        callback.model = model
        callback.on_epoch_end(epoch=0, logs=None)


if __name__ == "__main__":
    unittest.main()



import collections
import logging
import re
import string
from typing import List

import numpy as np
import tensorflow as tf

from .dataset import QuestionAnsweringDataset, QuestionAnsweringExample


class BaseMetricForQuestionAnswering(tf.keras.callbacks.Callback):
    """Base metric for qa."""

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, limit=None, **kwargs):
        examples = QuestionAnsweringDataset.jsonl_to_examples(input_files, vocab_file, **kwargs)
        if limit is not None and limit > 0:
            examples = examples[:limit]
        return cls(examples=examples, **kwargs)

    def __init__(self, examples: List[QuestionAnsweringExample], batch_size=32, **kwargs):
        super().__init__()
        self.examples = examples
        self.batch_size = batch_size

    def on_epoch_end(self, epoch, logs):
        input_ids, segment_ids, attention_mask = self._build_inputs()
        dataset = tf.data.Dataset.zip(
            (
                tf.data.Dataset.from_tensor_slices(input_ids),
                tf.data.Dataset.from_tensor_slices(segment_ids),
                tf.data.Dataset.from_tensor_slices(attention_mask),
            )
        ).batch(self.batch_size)
        dataset = dataset.map(lambda a, b, c: {"input_ids": a, "segment_ids": b, "attention_mask": c})
        outputs = self.model.predict(dataset)
        heads, tails = outputs[0], outputs[1]
        pred_answers, gold_answers = [], []
        for head, tail, example in zip(heads, tails, self.examples):
            head, tail = np.argmax(head), np.argmax(tail)
            pred_tokens = example.tokens[head : tail + 1]
            pred_text = "".join([str(x).lstrip("##") for x in pred_tokens])
            gold_text = "".join([str(x).lstrip("##") for x in example.tokens[example.start : example.end + 1]])
            pred_answers.append(self._normalize_answer(pred_text))
            gold_answers.append(self._normalize_answer(gold_text))
        self._compute_metric(epoch, pred_answers, gold_answers)

    def _compute_metric(self, epoch, pred_answers, gold_answers):
        raise NotImplementedError()

    def _build_inputs(self):
        input_ids, segment_ids, attention_mask = [], [], []
        maxlen = max([len(e.input_ids) for e in self.examples])
        for e in self.examples:
            _input_ids = e.input_ids + [0] * (maxlen - len(e.input_ids))
            _segment_ids = e.segment_ids + [0] * (maxlen - len(e.segment_ids))
            _attention_mask = e.attention_mask + [0] * (maxlen - len(e.attention_mask))
            input_ids.append(_input_ids)
            segment_ids.append(_segment_ids)
            attention_mask.append(_attention_mask)
        input_ids = tf.constant(input_ids, shape=(len(input_ids), maxlen), dtype=tf.int32)
        segment_ids = tf.constant(segment_ids, shape=(len(segment_ids), maxlen), dtype=tf.int32)
        attention_mask = tf.constant(attention_mask, shape=(len(attention_mask), maxlen), dtype=tf.int32)
        return input_ids, segment_ids, attention_mask

    def _normalize_answer(self, s):
        if not s:
            return ""
        # Lower text and remove punctuation, articles and extra whitespace.
        def remove_articles(text):
            regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
            return re.sub(regex, " ", text)

        def white_space_fix(text):
            return " ".join(text.split())

        def remove_punc(text):
            exclude = set(string.punctuation)
            return "".join(ch for ch in text if ch not in exclude)

        def lower(text):
            return text.lower()

        return white_space_fix(remove_articles(remove_punc(lower(s))))


class ExactMatchForQuestionAnswering(BaseMetricForQuestionAnswering):
    """Exact Match metric for question answering."""

    def _compute_metric(self, epoch, pred_answers, gold_answers):
        num_matchs = 0
        for pred, gold in zip(pred_answers, gold_answers):
            num_matchs += pred == gold
        acc = num_matchs * 1.0 / len(pred_answers)
        tf.summary.scalar("EM", acc, step=epoch, description="Exact Match Score")
        logging.info("No.%4d epoch EM: %.4f", epoch + 1, acc)


class F1ForQuestionAnswering(BaseMetricForQuestionAnswering):
    """F1 metric for question answering."""

    def _compute_metric(self, epoch, pred_answers, gold_answers):
        f1_scores = []
        for pred, gold in zip(pred_answers, gold_answers):
            pred, gold = self._normalize_answer(pred), self._normalize_answer(gold)
            f1_scores.append(self._compute_f1(pred, gold))
        f1 = sum(f1_scores) / len(f1_scores)
        tf.summary.scalar("F1", f1, step=epoch, description="F1 Score")
        logging.info("No.%4d epoch F1: %.4f", epoch + 1, f1)

    def _compute_f1(self, pred, gold):
        pred_toks = pred.split()
        gold_toks = gold.split()
        common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
        num_same = sum(common.values())
        if len(gold_toks) == 0 or len(pred_toks) == 0:
            # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
            return int(gold_toks == pred_toks)
        if num_same == 0:
            return 0
        precision = 1.0 * num_same / len(pred_toks)
        recall = 1.0 * num_same / len(gold_toks)
        f1 = (2 * precision * recall) / (precision + recall)
        return f1



import json
import logging
import re
from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

from .tokenizer import TokenClassificationLabelTokenizer

TokenClassificationExample = namedtuple(
    "TokenClassificationExample", ["tokens", "labels", "input_ids", "segment_ids", "attention_mask", "label_ids"]
)


class TokenClassificationDataset:
    """Dataset for token classification."""

    @classmethod
    def from_tfrecord_files(
        cls,
        input_files,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        **kwargs
    ):
        dataset = cls._read_tfrecord(input_files, **kwargs)
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def from_conll_files(cls, input_files, vocab_file, label_vocab_file, **kwargs):
        examples = cls.conll_to_examples(input_files, vocab_file, label_vocab_file, **kwargs)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, label_vocab_file, **kwargs):
        examples = cls.jsonl_to_examples(input_files, vocab_file, label_vocab_file, **kwargs)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls,
        examples: List[TokenClassificationExample],
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        logging.info("Number of examples: %d", len(examples))
        cls._show_examples(examples, n=5, verbose=verbose, **kwargs)
        dataset = cls._zip_dataset(examples)
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            dorp_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, verbose=True, **kwargs):
        if not verbose:
            return
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("No.%d example: %s", i, examples[i])

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        """zip dataset"""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.label_ids for e in examples]),
            )
        )
        return dataset

    @classmethod
    def _bucketing(
        cls,
        dataset,
        batch_size=64,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_size) should equals len(bucket_doundaries) + 1"

        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.apply(tf.data.experimental.bucket_by_sequence_length(
            element_length_func=lambda a, b, c, y: tf.size(a),
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            padded_shapes=([None, ], [None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )).prefetch(tf.data.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset):
        dataset = dataset.map(
            lambda a, b, c, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, y),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _read_tfrecord(cls, input_files, **kwargs):
        dataset = tf.data.Dataset.from_tensor_slices(input_files)
        dataset = dataset.interleave(
            lambda x: tf.data.TFRecordDataset(x),
            cycle_length=len(input_files),
            num_parallel_calls=tf.data.AUTOTUNE,
        )
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "label_ids": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["label_ids"]), tf.int32),
            ),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def conll_to_tfrecord(cls, input_files, vocab_file, label_vocab_file, output_files, **kwargs):
        examples = cls.conll_to_examples(input_files, vocab_file, label_vocab_file, **kwargs)
        cls.examples_to_tfrecord(examples, output_files, **kwargs)

    @classmethod
    def conll_to_examples(cls, input_files, vocab_file, label_vocab_file, sep="\t", **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        instances = []
        for f in input_files:
            features, labels = [], []
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        instances.append({"features": features, "labels": labels})
                        features, labels = [], []
                        continue
                    parts = re.split(sep, line)
                    if len(parts) != 2:
                        continue
                    features.append(parts[0])
                    labels.append(parts[1])
            if features and labels:
                instances.append({"features": features, "labels": labels})
        examples = []
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.pop("do_lower_case", True))
        label_tokenizer = TokenClassificationLabelTokenizer.from_file(
            label_vocab_file, o_token=kwargs.pop("o_token", "O")
        )
        for instance in instances:
            example = cls._instance_to_example(instance, tokenizer, label_tokenizer, **kwargs)
            if not example:
                continue
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def jsonl_to_tfrecord(cls, input_files, vocab_file, lable_vocab_file, output_files, **kwargs):
        examples = cls.jsonl_to_examples(input_files, vocab_file, lable_vocab_file, **kwargs)
        cls.examples_to_tfrecord(examples, output_files, **kwargs)

    @classmethod
    def jsonl_to_examples(cls, input_files, vocab_file, label_vocab_file, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        instances = []
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = json.loads(line)
                    instances.append(instance)
        logging.info("Collected %d instances in total.", len(instances))
        examples = []
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.pop("do_lower_case", True))
        label_tokenizer = TokenClassificationLabelTokenizer.from_file(
            label_vocab_file, o_token=kwargs.pop("o_token", "O")
        )
        for instance in instances:
            example = cls._instance_to_example(instance, tokenizer, label_tokenizer, **kwargs)
            if not example:
                continue
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def examples_to_tfrecord(cls, examples, output_files, **kwargs):
        if isinstance(output_files, str):
            output_files = [output_files]
        writers = [tf.io.TFRecordWriter(f) for f in output_files]
        idx = 0
        for example in examples:
            tfrecord_example = cls._example_to_tfrecord(example)
            writers[idx].write(tfrecord_example.SerializeToString())
            idx += 1
            idx = idx % len(writers)
        for w in writers:
            w.close()
        logging.info("Finished to write %d examples to tfrecords.", len(examples))

    @classmethod
    def _int64_feature(cls, values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=values))

    @classmethod
    def _instance_to_example(
        cls, instance, tokenizer: BertWordPieceTokenizer, label_tokenizer: TokenClassificationLabelTokenizer, **kwargs
    ):
        features = instance[kwargs.get("features_key", "features")]
        tokens = ["[CLS]"] + features + ["[SEP]"]
        cls_id, sep_id = tokenizer.token_to_id("[CLS]"), tokenizer.token_to_id("[SEP]")
        input_ids = [cls_id] + [tokenizer.token_to_id(token) for token in features] + [sep_id]
        labels = instance[kwargs.get("labels_key", "labels")]
        label_ids = label_tokenizer.labels_to_ids(labels, add_cls=True, add_sep=True)
        example = TokenClassificationExample(
            tokens=tokens,
            labels=labels,
            input_ids=input_ids,
            segment_ids=[0] * len(input_ids),
            attention_mask=[1] * len(input_ids),
            label_ids=label_ids,
        )
        return example

    @classmethod
    def _example_to_tfrecord(cls, example: TokenClassificationExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "label_ids": cls._int64_feature([int(x) for x in example.label_ids]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))



import logging

import tensorflow as tf

from transformers_keras.modeling_albert import (
    Albert,
    AlbertEmbedding,
    AlbertEncoder,
    AlbertEncoderGroup,
    AlbertEncoderLayer,
    AlbertModel,
    AlbertPretrainedModel,
)
from transformers_keras.modeling_bert import (
    Bert,
    BertEmbedding,
    BertEncoder,
    BertEncoderLayer,
    BertIntermediate,
    BertModel,
    BertPretrainedModel,
)
from transformers_keras.modeling_utils import complete_inputs
from transformers_keras.question_answering.callback import (
    BaseMetricForQuestionAnswering,
    ExactMatchForQuestionAnswering,
    F1ForQuestionAnswering,
)
from transformers_keras.question_answering.dataset import (
    QuestionAnsweringDataset,
    QuestionAnsweringExample,
    QuestionAnsweringXDataset,
    QuestionAnsweringXExample,
)
from transformers_keras.question_answering.models import (
    AlbertForQuestionAnswering,
    AlbertForQuestionAnsweringX,
    BertForQuestionAnswering,
    BertForQuestionAnsweringX,
)
from transformers_keras.sentence_embedding.bert_embedding import BertForSentenceEmbedding
from transformers_keras.sentence_embedding.callback import SpearmanForSentenceEmbedding
from transformers_keras.sentence_embedding.simcse_dataset import SimCSEDataset, SimCSEExample
from transformers_keras.sentence_embedding.simcse_models import HardNegativeSimCSE, SupervisedSimCSE, UnsupervisedSimCSE
from transformers_keras.sequence_classification.dataset import (
    SequenceClassificationDataset,
    SequenceClassificationExample,
)
from transformers_keras.sequence_classification.models import (
    AlbertForSequenceClassification,
    BertForSequenceClassification,
)
from transformers_keras.token_classification.callback import SeqEvalForTokenClassification
from transformers_keras.token_classification.crf_models import (
    AlertCRFForTokenClassification,
    BertCRFForTokenClassification,
    CRFModel,
)
from transformers_keras.token_classification.dataset import TokenClassificationDataset, TokenClassificationExample
from transformers_keras.token_classification.models import AlbertForTokenClassification, BertForTokenClassification

__name__ = "transformers_keras"
__version__ = "0.4.4"

logging.basicConfig(format="%(asctime)s %(levelname)7s %(filename)20s %(lineno)4d] %(message)s", level=logging.INFO)



import tensorflow as tf
from transformers_keras.modeling_albert import AlbertModel, AlbertPretrainedModel
from transformers_keras.modeling_bert import BertModel, BertPretrainedModel


class BertForSequenceClassification(BertPretrainedModel):
    """Bert for sequence classification"""

    def __init__(
        self,
        num_labels=2,
        vocab_size=21128,
        max_positions=512,
        hidden_size=768,
        type_vocab_size=2,
        num_layers=6,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        initializer_range=0.02,
        epsilon=1e-12,
        **kwargs
    ):
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        bert_model = BertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            hidden_size=hidden_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="bert",
        )

        _, pooled_output, _, _ = bert_model(input_ids, segment_ids, attention_mask)
        logits = tf.keras.layers.Dense(num_labels, name="logits")(pooled_output)
        super().__init__(inputs=[input_ids, segment_ids, attention_mask], outputs=[logits], **kwargs)

        self.bert_model = bert_model
        self.num_labels = num_labels


class AlbertForSequenceClassification(AlbertPretrainedModel):
    """Albert for sequence classification."""

    def __init__(
        self,
        num_labels=2,
        vocab_size=21128,
        max_positions=512,
        embedding_size=128,
        type_vocab_size=2,
        num_layers=12,
        num_groups=1,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        **kwargs
    ):
        # build functional model
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        albert_model = AlbertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            embedding_size=embedding_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_groups=num_groups,
            num_layers_each_group=num_layers_each_group,
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            epsilon=epsilon,
            initializer_range=initializer_range,
            name="albert",
        )

        _, pooled_output, _, _ = albert_model(input_ids, segment_ids, attention_mask)
        logits = tf.keras.layers.Dense(num_labels, name="logits")(pooled_output)
        super().__init__(inputs=[input_ids, segment_ids, attention_mask], outputs=[logits], **kwargs)

        self.num_labels = num_labels
        self.albert_model = albert_model



import json
import unittest

from transformers_keras.question_answering.dataset import QuestionAnsweringDataset, QuestionAnsweringXDataset


class DatasetTest(unittest.TestCase):
    """Dataset test."""

    def _read_qa_instances(self, input_file):
        instances = []
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                data = json.loads(line)
                instances.append(
                    {
                        "context": data["passage"],
                        "question": data["question"],
                        "answer": data["answer"],
                    }
                )
        return instances

    def _read_qax_instances(self, input_file):
        instances = []
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                data = json.loads(line)
                instances.append(
                    {
                        "context": data["passage"],
                        "question": data["question"],
                        "answer": data["answer"],
                        "class": data["class"],
                    }
                )
        return instances

    def test_qa_dataset(self):
        instances = self._read_qa_instances("testdata/qa.sogouqa.jsonl")
        # test jsonl to examples
        examples = QuestionAnsweringDataset.jsonl_to_examples(
            ["testdata/qa.sogouqa.jsonl"], "testdata/vocab.bert.txt", context_key="passage"
        )
        for idx, e in enumerate(examples):
            print("{} -> {}".format(instances[idx]["answer"], " ".join(e.tokens[e.start : e.end + 1])))

        # test examples to tfrecord
        QuestionAnsweringDataset.examples_to_tfrecord(examples, "testdata/qa.sogouqa.tfrecord")
        d = QuestionAnsweringDataset.from_tfrecord_files(
            ["testdata/qa.sogouqa.tfrecord"] * 4,
            batch_size=4,
        )
        print()
        print(next(iter(d)))

        # test jsonl to tfrecord
        QuestionAnsweringDataset.jsonl_to_tfrecord(
            "testdata/qa.sogouqa.jsonl",
            "testdata/vocab.bert.txt",
            "testdata/qa.sogouqa.tfrecord",
            context_key="passage",
        )

        # test from tfrecord
        d = QuestionAnsweringDataset.from_tfrecord_files("testdata/qa.sogouqa.tfrecord", batch_size=4)
        print()
        print(next(iter(d)))

        # test from jsonl
        d = QuestionAnsweringDataset.from_jsonl_files(
            "testdata/qa.sogouqa.jsonl", "testdata/vocab.bert.txt", context_key="passage", batch_size=4
        )
        print()
        print(next(iter(d)))

    def test_qax_dataset(self):
        instances = self._read_qax_instances("testdata/qax.sogouqa.jsonl")
        # test jsonl to examples
        examples = QuestionAnsweringXDataset.jsonl_to_examples(
            ["testdata/qax.sogouqa.jsonl"], "testdata/vocab.bert.txt", context_key="passage"
        )
        for idx, e in enumerate(examples):
            print("{} -> {}".format(instances[idx]["answer"], " ".join(e.tokens[e.start : e.end + 1])))

        # test examples to tfrecord
        QuestionAnsweringXDataset.examples_to_tfrecord(examples, "testdata/qax.sogouqa.tfrecord")
        d = QuestionAnsweringXDataset.from_tfrecord_files(
            ["testdata/qax.sogouqa.tfrecord"] * 4,
            batch_size=4,
        )
        print()
        print(next(iter(d)))

        # test jsonl to tfrecord
        QuestionAnsweringXDataset.jsonl_to_tfrecord(
            "testdata/qax.sogouqa.jsonl",
            "testdata/vocab.bert.txt",
            "testdata/qax.sogouqa.tfrecord",
            context_key="passage",
        )

        # test from tfrecord
        d = QuestionAnsweringXDataset.from_tfrecord_files("testdata/qax.sogouqa.tfrecord", batch_size=4)
        print()
        print(next(iter(d)))

        # test from jsonl
        d = QuestionAnsweringXDataset.from_jsonl_files(
            "testdata/qax.sogouqa.jsonl", "testdata/vocab.bert.txt", context_key="passage", batch_size=4
        )
        print()
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
import tensorflow_addons as tfa
from transformers_keras.modeling_albert import (AlbertModel,
                                                AlbertPretrainedModel)
from transformers_keras.modeling_bert import BertModel, BertPretrainedModel


def _unpack_data(data):
    """Support sample_weight"""
    if len(data) == 3:
        x, y, sample_weight = data
    elif len(data) == 2:
        x, y, sample_weight = data[0], data[1], None
    elif len(data) == 1:
        x, y, sample_weight = data[0], None, None
    return x, y, sample_weight


class CRFModel(tf.keras.Model):
    """Wrappe a CRF layer for base model."""

    def __init__(
        self,
        model: tf.keras.Model,
        units: int,
        chain_initializer="orthogonal",
        use_boundary: bool = True,
        boundary_initializer="zeros",
        use_kernel: bool = True,
        **kwargs
    ):
        # build functional model
        crf = tfa.layers.CRF(
            units=units,
            chain_initializer=chain_initializer,
            use_boundary=use_boundary,
            boundary_initializer=boundary_initializer,
            use_kernel=use_kernel,
            **kwargs
        )
        # take model's first output passed to CRF layer
        decode_sequence, potentials, sequence_length, kernel = crf(inputs=model.outputs[0])
        # set name for outputs
        decode_sequence = tf.keras.layers.Lambda(lambda x: x, name="decode_sequence")(decode_sequence)
        potentials = tf.keras.layers.Lambda(lambda x: x, name="potentials")(potentials)
        sequence_length = tf.keras.layers.Lambda(lambda x: x, name="sequence_length")(sequence_length)
        kernel = tf.keras.layers.Lambda(lambda x: x, name="kernel")(kernel)
        super().__init__(inputs=model.inputs, outputs=[decode_sequence, potentials, sequence_length, kernel], **kwargs)
        self.crf = crf

    def train_step(self, data):
        x, y, sample_weight = _unpack_data(data)
        with tf.GradientTape() as tape:
            decode_sequence, potentials, sequence_length, kernel = self(x, training=True)
            crf_loss = -tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel)[0]
            if sample_weight is not None:
                crf_loss = crf_loss * sample_weight
            crf_loss = tf.reduce_mean(crf_loss)
            loss = crf_loss + sum(self.losses)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))

        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def test_step(self, data):
        x, y, sample_weight = _unpack_data(data)
        decode_sequence, potentials, sequence_length, kernel = self(x, training=False)
        crf_loss = -tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel)[0]
        if sample_weight is not None:
            crf_loss = crf_loss * sample_weight
        crf_loss = tf.reduce_mean(crf_loss)
        loss = crf_loss + sum(self.losses)
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def predict_step(self, data):
        x, _, _ = _unpack_data(data)
        decode_sequence, potentials, sequence_length, kernel = self(x, training=False)
        return decode_sequence


class BertCRFForTokenClassification(BertPretrainedModel):
    """BERT+CRF use for token classification."""

    def __init__(
        self,
        num_labels,
        vocab_size=21128,
        max_positions=512,
        hidden_size=768,
        type_vocab_size=2,
        num_layers=6,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        initializer_range=0.02,
        epsilon=1e-12,
        **kwargs
    ):
        # build functional model
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        bert_model = BertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            hidden_size=hidden_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="bert",
        )
        crf = tfa.layers.CRF(num_labels)
        sequence_outputs, _, _, _ = bert_model(input_ids, segment_ids, attention_mask)
        decode_sequence, potentials, sequence_length, kernel = crf(sequence_outputs)
        super().__init__(
            inputs=[input_ids, segment_ids, attention_mask],
            outputs=[decode_sequence, potentials, sequence_length, kernel],
            **kwargs
        )

        self.num_labels = num_labels
        self.bert_model = bert_model
        self.crf = crf

    @tf.function( # fmt: skip
        input_signature=[{
            "input_ids": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="input_ids"), # fmt: skip
            "segment_ids": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="segment_ids"), # fmt: skip
            "attention_mask": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="attention_mask"), # fmt: skip
        }]
    )
    def forward(self, inputs):
        input_ids, segment_ids, attention_mask = inputs["input_ids"], inputs["segment_ids"], inputs["attention_mask"]
        sequence, _, _, _ = self(inputs=[input_ids, segment_ids, attention_mask], training=False)
        return {"decode_sequence": sequence}

    def train_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        with tf.GradientTape() as tape:
            decode_sequence, potentials, sequence_length, kernel = self(
                inputs=[input_ids, segment_ids, attention_mask], training=True
            )
            crf_loss = -tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel)[0]
            crf_loss = tf.reduce_mean(crf_loss)
            loss = crf_loss + sum(self.losses)

        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def test_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        decode_sequence, potentials, sequence_length, kernel = self(
            inputs=[input_ids, segment_ids, attention_mask], training=False
        )
        crf_loss = tf.reduce_mean(-tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel))
        loss = crf_loss + sum(self.losses)
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def predict_step(self, data):
        x, _, _ = _unpack_data(data)
        decode_sequence, potentials, sequence_length, kernel = self(x, training=False)
        return decode_sequence



class AlertCRFForTokenClassification(AlbertPretrainedModel):
    """ALBERT+CRF use for token classification."""

    def __init__(
        self,
        num_labels,
        vocab_size=21128,
        max_positions=512,
        embedding_size=128,
        type_vocab_size=2,
        num_layers=12,
        num_groups=1,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        **kwargs
    ):
        # build functional model
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        albert_model = AlbertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            embedding_size=embedding_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_groups=num_groups,
            num_layers_each_group=num_layers_each_group,
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            epsilon=epsilon,
            initializer_range=initializer_range,
            name="albert",
        )
        crf = tfa.layers.CRF(num_labels)
        sequence_outputs, _, _, _ = albert_model(input_ids, segment_ids, attention_mask)
        decode_sequence, potentials, sequence_length, kernel = crf(sequence_outputs)
        super().__init__(
            inputs=[input_ids, segment_ids, attention_mask],
            outputs=[decode_sequence, potentials, sequence_length, kernel],
            **kwargs
        )

        self.num_labels = num_labels
        self.albert_model = albert_model
        self.crf = crf

    @tf.function(
        input_signature=[{
            "input_ids": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="input_ids"), # fmt: skip
            "segment_ids": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="segment_ids"), # fmt: skip
            "attention_mask": tf.TensorSpec(shape=(None, None,), dtype=tf.int32, name="attention_mask"), # fmt: skip
        }]
    )
    def forward(self, inputs):
        input_ids, segment_ids, attention_mask = inputs["input_ids"], inputs["segment_ids"], inputs["attention_mask"]
        sequence, _, _, _ = self(inputs=[input_ids, segment_ids, attention_mask], training=False)
        return {"decode_sequence": sequence}

    def train_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        with tf.GradientTape() as tape:
            decode_sequence, potentials, sequence_length, kernel = self(
                inputs=[input_ids, segment_ids, attention_mask], training=True
            )
            crf_loss = -tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel)[0]
            crf_loss = tf.reduce_mean(crf_loss)
            loss = crf_loss + sum(self.losses)

        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def test_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        decode_sequence, potentials, sequence_length, kernel = self(
            inputs=[input_ids, segment_ids, attention_mask], training=False
        )
        crf_loss = tf.reduce_mean(-tfa.text.crf_log_likelihood(potentials, y, sequence_length, kernel))
        loss = crf_loss + sum(self.losses)
        self.compiled_metrics.update_state(y, potentials)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results
        
    def predict_step(self, data):
        x, _, _ = _unpack_data(data)
        decode_sequence, potentials, sequence_length, kernel = self(x, training=False)
        return decode_sequence




import json
import logging
import re
from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

QuestionAnsweringExample = namedtuple(
    "QuestionAnsweringExample", ["text", "tokens", "input_ids", "segment_ids", "attention_mask", "start", "end"]
)
QuestionAnsweringXExample = namedtuple(
    "QuestionAnsweringXExample",
    ["text", "tokens", "input_ids", "segment_ids", "attention_mask", "start", "end", "class_id"],
)


class _BaseQuestionAnsweringDataset:
    """Dataset transformation"""

    @classmethod
    def from_jsonl_files(
        cls,
        input_files,
        vocab_file,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        examples = cls.jsonl_to_examples(input_files, vocab_file, **kwargs)
        return cls.from_examples(
            examples,
            batch_size=batch_size,
            repeat=repeat,
            max_sequence_length=max_sequence_length,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            buffer_size=buffer_size,
            seed=seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
            pad_id=pad_id,
            auto_shard_policy=auto_shard_policy,
            drop_remainder=drop_remainder,
            verbose=verbose,
            **kwargs,
        )

    @classmethod
    def from_examples(
        cls,
        examples,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        raise NotImplementedError()

    @classmethod
    def jsonl_to_tfrecord(cls, input_files, vocab_file, output_files, **kwargs):
        examples = cls.jsonl_to_examples(input_files, vocab_file, **kwargs)
        cls.examples_to_tfrecord(examples, output_files, **kwargs)

    @classmethod
    def jsonl_to_examples(cls, input_files, vocab_file, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        instances = []
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = json.loads(line)
                    instances.append(instance)
        logging.info("Collected %d instances in total.", len(instances))
        examples = []
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.pop("do_lower_case", True))
        for instance in instances:
            example = cls._instance_to_example(instance, tokenizer, **kwargs)
            if not example:
                continue
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def _instance_to_example(cls, instance, tokenizer: BertWordPieceTokenizer, **kwargs):
        raise NotImplementedError()

    @classmethod
    def examples_to_tfrecord(cls, examples, output_files, **kwargs):
        if isinstance(output_files, str):
            output_files = [output_files]
        writers = [tf.io.TFRecordWriter(f) for f in output_files]
        idx = 0
        for example in examples:
            tfrecord_example = cls._example_to_tfrecord(example)
            writers[idx].write(tfrecord_example.SerializeToString())
            idx += 1
            idx = idx % len(writers)
        for w in writers:
            w.close()
        logging.info("Finished to write %d examples to tfrecords.", len(examples))

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _find_answer_span(cls, context, answer):
        for m in re.finditer(re.escape(answer), context, re.IGNORECASE):
            start, end = m.span()
            return start, end
        return 0, 0

    @classmethod
    def _int64_feature(cls, values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


class QuestionAnsweringDataset(_BaseQuestionAnsweringDataset):
    """Dataset builder for question answering models."""

    @classmethod
    def from_tfrecord_files(
        cls,
        input_files,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        drop_remainder=False,
        auto_shard_policy=None,
        **kwargs
    ):
        dataset = cls._read_tfrecord(input_files, **kwargs)
        dataset = dataset.filter(lambda a, b, c, x, y: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def from_examples(
        cls,
        examples: List[QuestionAnsweringExample],
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        logging.info("Load %d examples in total.", len(examples))
        if verbose:
            cls._show_examples(examples, n=5, **kwargs)
        dataset = cls._zip_dataset(examples)
        dataset = dataset.filter(
            lambda a, b, c, x, y: tf.size(a) <= max_sequence_length,
        )
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, **kwargs):
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("NO.%d example: %s", i, examples[i])

    @classmethod
    def _zip_dataset(cls, examples: List[QuestionAnsweringExample]):
        """Zip datasets to one dataset."""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _bucketing(
        cls,
        dataset,
        batch_size=64,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_sizes) should equals len(bucket_doundaries) + 1"

        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.apply(tf.data.experimental.bucket_by_sequence_length(
            element_length_func=lambda a, b, c, x, y: tf.size(a),
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            padded_shapes=([None, ], [None, ], [None, ], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None),
            drop_remainder=drop_remainder,
        )).prefetch(tf.data.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset):
        dataset = dataset.map(
            lambda a, b, c, x, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, {"head": x, "tail": y}),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _read_tfrecord(cls, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        if len(input_files) == 1:
            dataset = tf.data.TFRecordDataset(input_files)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(input_files)
            dataset = dataset.interleave(
                lambda x: tf.data.TFRecordDataset(x),
                cycle_length=len(input_files),
                num_parallel_calls=tf.data.AUTOTUNE,
            )
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["start"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["end"])), tf.int32),
            ),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: QuestionAnsweringExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "start": cls._int64_feature([int(example.start)]),
            "end": cls._int64_feature([int(example.end)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _instance_to_example(cls, instance, tokenizer: BertWordPieceTokenizer, **kwargs):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        answer = instance[kwargs.get("answer_key", "answer")]
        start_char_idx, end_char_idx = cls._find_answer_span(context, answer)
        if end_char_idx <= start_char_idx:
            return None
        # Mark the character indexes in context that are in answer
        is_char_in_ans = [0] * len(context)
        for idx in range(start_char_idx, end_char_idx):
            is_char_in_ans[idx] = 1
        context_encoding = tokenizer.encode(context)
        # Find tokens that were created from answer characters
        ans_token_idx = []
        for idx, (start_char_idx, end_char_idx) in enumerate(context_encoding.offsets):
            if sum(is_char_in_ans[start_char_idx:end_char_idx]) > 0:
                ans_token_idx.append(idx)
        if not ans_token_idx:
            return None
        start_token_idx, end_token_idx = ans_token_idx[0], ans_token_idx[-1]
        question_encoding = tokenizer.encode(question)
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(question_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        assert len(input_ids) == len(segment_ids), "input_ids length:{} VS segment_ids length: {}".format(
            len(input_ids), len(segment_ids)
        )
        assert len(input_ids) == len(attention_mask), "input_ids length:{} VS attention_mask length: {}".format(
            len(input_ids), len(attention_mask)
        )

        example = QuestionAnsweringExample(
            text=context + question,
            tokens=context_encoding.tokens + question_encoding.tokens[1:],
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=start_token_idx,
            end=end_token_idx,
        )
        return example


class QuestionAnsweringXDataset(_BaseQuestionAnsweringDataset):
    """Dataset builder for question answering models."""

    @classmethod
    def from_tfrecord_files(
        cls,
        input_files,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        drop_remainder=False,
        auto_shard_policy=None,
        **kwargs
    ):
        dataset = cls._read_tfrecord(input_files, **kwargs)
        dataset = dataset.filter(lambda a, b, c, x, y, z: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def from_examples(
        cls,
        examples: List[QuestionAnsweringXExample],
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        logging.info("Load %d examples in total.", len(examples))
        if verbose:
            cls._show_examples(examples, n=5, **kwargs)
        dataset = cls._zip_dataset(examples)
        dataset = dataset.filter(
            lambda a, b, c, x, y, z: tf.size(a) <= max_sequence_length,
        )
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, **kwargs):
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("NO.%d example: %s", i, examples[i])

    @classmethod
    def _zip_dataset(cls, examples: List[QuestionAnsweringXExample]):
        """Zip datasets to one dataset."""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.class_id for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _bucketing(
        cls,
        dataset,
        batch_size=64,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_sizes) should equals len(bucket_doundaries) + 1"

        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.apply(tf.data.experimental.bucket_by_sequence_length(
            element_length_func=lambda a, b, c, x, y, z: tf.size(a),
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            padded_shapes=([None, ], [None, ], [None, ], [], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None, None),
            drop_remainder=drop_remainder,
        )).prefetch(tf.data.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset):
        dataset = dataset.map(
            lambda a, b, c, x, y, z: (
                {"input_ids": a, "segment_ids": b, "attention_mask": c},
                {"head": x, "tail": y, "class": z},
            ),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _read_tfrecord(cls, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        if len(input_files) == 1:
            dataset = tf.data.TFRecordDataset(input_files)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(input_files)
            dataset = dataset.interleave(
                lambda x: tf.data.TFRecordDataset(x),
                cycle_length=len(input_files),
                num_parallel_calls=tf.data.AUTOTUNE,
            )
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
            "class_id": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["start"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["end"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["class_id"])), tf.int32),
            ),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: QuestionAnsweringXExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "start": cls._int64_feature([int(example.start)]),
            "end": cls._int64_feature([int(example.end)]),
            "class_id": cls._int64_feature([int(example.class_id)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _instance_to_example(cls, instance, tokenizer: BertWordPieceTokenizer, **kwargs):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        answer = instance[kwargs.get("answer_key", "answer")]
        context_encoding = tokenizer.encode(context)
        question_encoding = tokenizer.encode(question)

        start_char_idx, end_char_idx = cls._find_answer_span(context, answer)
        # invalid answer index
        if end_char_idx <= start_char_idx:
            return cls._invalid_answer_example(instance, context_encoding, question_encoding, **kwargs)

        # Mark the character indexes in context that are in answer
        is_char_in_ans = [0] * len(context)
        for idx in range(start_char_idx, end_char_idx):
            is_char_in_ans[idx] = 1
        # Find tokens that were created from answer characters
        ans_token_idx = []
        for idx, (start_char_idx, end_char_idx) in enumerate(context_encoding.offsets):
            if sum(is_char_in_ans[start_char_idx:end_char_idx]) > 0:
                ans_token_idx.append(idx)
        # no answer tokens found
        if not ans_token_idx:
            return cls._invalid_answer_example(instance, context_encoding, question_encoding, **kwargs)

        start_token_idx, end_token_idx = ans_token_idx[0], ans_token_idx[-1]
        return cls._valid_answer_example(
            instance, context_encoding, question_encoding, start_token_idx, end_token_idx, **kwargs
        )

    @classmethod
    def _invalid_answer_example(cls, instance, context_encoding, question_encoding, **kwargs):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        class_id = instance[kwargs.get("class_key", "class")]
        # build inputs
        tokens = context_encoding.tokens + question_encoding.tokens[1:]
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(question_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        return QuestionAnsweringXExample(
            text=context + question,
            tokens=tokens,
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=0,
            end=0,
            class_id=class_id,
        )

    @classmethod
    def _valid_answer_example(
        cls, instance, context_encoding, question_encoding, start_token_idx, end_token_idx, **kwargs
    ):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        class_id = instance[kwargs.get("class_key", "class")]
        # build inputs
        tokens = context_encoding.tokens + question_encoding.tokens[1:]
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(context_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        return QuestionAnsweringXExample(
            text=context + question,
            tokens=tokens,
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=start_token_idx,
            end=end_token_idx,
            class_id=class_id,
        )



import logging
import os
import re
import unittest

from tokenizers import BertWordPieceTokenizer
from transformers_keras.sentence_embedding.callback import ExampleForSpearman, SpearmanForSentenceEmbedding
from transformers_keras.sentence_embedding.simcse_models import HardNegativeSimCSE, SupervisedSimCSE, UnsupervisedSimCSE

BERT_PATH = os.path.join(os.environ["CHINESE_BERT_PATH"], "chinese_roberta_wwm_ext_L-12_H-768_A-12")
VOCAB_PATH = os.path.join(BERT_PATH, "vocab.txt")


class CallbackTest(unittest.TestCase):
    """Callback test"""

    def _read_examples(self):
        examples = []
        p = os.path.join(os.environ["CHINESE_STS_PATH"], "cnsd-sts-test.txt")
        with open(p, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                parts = line.split("||")
                if len(parts) != 4:
                    continue
                examples.append(ExampleForSpearman(sentence_a=parts[1], sentence_b=parts[2], label=int(parts[3])))
        logging.info("Load %d examples.", len(examples))
        return examples

    def test_spearman_for_sentence_embedding(self):
        examples = self._read_examples()
        tokenizer = BertWordPieceTokenizer.from_file(VOCAB_PATH, lowercase=True)
        callback = SpearmanForSentenceEmbedding(examples, tokenizer)

        def _compute_spearman(model):
            callback.model = model
            callback.on_epoch_end(epoch=0, logs=None)

        _compute_spearman(HardNegativeSimCSE.from_pretrained(BERT_PATH))
        _compute_spearman(SupervisedSimCSE.from_pretrained(BERT_PATH))
        _compute_spearman(UnsupervisedSimCSE.from_pretrained(BERT_PATH))


if __name__ == "__main__":
    unittest.main()



import json
import logging
import os
from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

SequenceClassificationExample = namedtuple(
    "SequenceClassificationExample", ["tokens", "input_ids", "segment_ids", "attention_mask", "label"]
)


class SequenceClassificationDataset:
    """Dataset builder for sequence classification."""

    @classmethod
    def from_tfrecord_files(
        cls,
        input_files,
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        **kwargs
    ):
        dataset = cls._read_tfrecord(input_files, **kwargs)
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, **kwargs):
        examples = cls.jsonl_to_examples(input_files, vocab_file, **kwargs)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls,
        examples: List[SequenceClassificationExample],
        batch_size=64,
        repeat=None,
        max_sequence_length=512,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        verbose=True,
        **kwargs
    ):
        logging.info("Number of examples: %d", len(examples))
        cls._show_examples(examples, n=5, verbose=verbose, **kwargs)
        dataset = cls._zip_dataset(examples)
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        if repeat is not None:
            dataset = dataset.repeat(repeat)
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        dataset = cls._bucketing(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs
        )
        dataset = cls._to_dict(dataset)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, verbose=True, **kwargs):
        if not verbose:
            return
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("No.%d example: %s", i, examples[i])

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        """zip dataset"""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.label for e in examples]),
            )
        )
        return dataset

    @classmethod
    def _bucketing(
        cls,
        dataset,
        batch_size=64,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_sizes) should equals len(bucket_doundaries) + 1"

        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.apply(tf.data.experimental.bucket_by_sequence_length(
            element_length_func=lambda a, b, c, y: tf.size(a),
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            padded_shapes=([None, ], [None, ], [None, ], []),
            padding_values=(pad_id, pad_id, pad_id, None),
            drop_remainder=drop_remainder,
        )).prefetch(tf.data.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset):
        dataset = dataset.map(
            lambda a, b, c, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, y),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _read_tfrecord(cls, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        if len(input_files) == 1:
            dataset = tf.data.TFRecordDataset(input_files)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(input_files)
            dataset = dataset.interleave(
                lambda x: tf.data.TFRecordDataset(x),
                cycle_length=len(input_files),
                num_parallel_calls=tf.data.AUTOTUNE,
            )
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "label": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["label"])), tf.int32),
            ),
            num_parallel_calls=tf.data.AUTOTUNE,
        ).prefetch(tf.data.AUTOTUNE)
        return dataset

    @classmethod
    def jsonl_to_tfrecord(cls, input_files, vocab_file, output_files, **kwargs):
        examples = cls.jsonl_to_examples(input_files, vocab_file, **kwargs)
        cls.examples_to_tfrecord(examples, output_files, **kwargs)
        logging.info("Convert data from jsonl to tfrecord finished.")

    @classmethod
    def jsonl_to_examples(cls, input_files, vocab_file, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        examples = []
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.get("do_lower_case", True))
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = json.loads(line)
                    sequence = instance[kwargs.get("sequence_key", "sequence")]
                    label = instance[kwargs.get("label_key", "label")]
                    encoding = tokenizer.encode(sequence)
                    examples.append(
                        SequenceClassificationExample(
                            tokens=encoding.tokens,
                            input_ids=encoding.ids,
                            segment_ids=encoding.type_ids,
                            attention_mask=encoding.attention_mask,
                            label=int(label),
                        )
                    )
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def examples_to_tfrecord(cls, examples, output_files, **kwargs):
        if isinstance(output_files, str):
            output_files = [output_files]
        writers = [tf.io.TFRecordWriter(f) for f in output_files]
        idx = 0
        for example in examples:
            tfrecord_example = cls._example_to_tfrecord(example)
            writers[idx].write(tfrecord_example.SerializeToString())
            idx += 1
            idx = idx % len(writers)
        for w in writers:
            w.close()

    @classmethod
    def _example_to_tfrecord(cls, example: SequenceClassificationExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "label": cls._int64_feature([int(example.label)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _int64_feature(cls, values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=values))



import os
import unittest

from transformers_keras.token_classification.models import AlbertForTokenClassification, BertForTokenClassification


class ModelsTest(unittest.TestCase):
    """Model tests"""

    def test_bert_for_token_classification(self):
        m = BertForTokenClassification(4)
        m.summary()
        m.save("models/bert-for-token-classification/1")

        bert_path = os.path.join(os.environ["CHINESE_BERT_PATH"], "chinese_roberta_wwm_ext_L-12_H-768_A-12")
        m = BertForTokenClassification.from_pretrained(bert_path, override_params={"num_labels": 4})
        m.summary()
        m.save("models/bert-for-token-classification/2")

    def test_albert_for_token_classification(self):
        m = AlbertForTokenClassification(4)
        m.summary()
        m.save("models/albert-for-token-classification/1")

        albert_path = os.path.join(os.environ["GOOGLE_ALBERT_PATH"], "albert-base-zh")
        m = AlbertForTokenClassification.from_pretrained(albert_path, override_params={"num_labels": 4})
        m.summary()
        m.save("models/albert-for-token-classification/2")


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from transformers_keras.question_answering.callback import ExactMatchForQuestionAnswering, F1ForQuestionAnswering
from transformers_keras.question_answering.models import BertForQuestionAnswering

BERT_PATH = os.path.join(os.environ["CHINESE_BERT_PATH"], "chinese_roberta_wwm_ext_L-12_H-768_A-12")
VOCAB_PATH = os.path.join(BERT_PATH, "vocab.txt")


class CallbackTest(unittest.TestCase):
    """Callback test"""

    def test_em_for_qa(self):
        callback = ExactMatchForQuestionAnswering.from_jsonl_files(
            input_files=[os.path.join(os.environ["SOGOUQA_PATH"], "sogouqa.test.jsonl")],
            vocab_file=VOCAB_PATH,
            limit=32,
            context_key="passage",
        )
        model = BertForQuestionAnswering.from_pretrained(BERT_PATH)
        callback.model = model
        callback.on_epoch_end(epoch=0, logs=None)

    def test_f1_for_qa(self):
        callback = F1ForQuestionAnswering.from_jsonl_files(
            input_files=[os.path.join(os.environ["SOGOUQA_PATH"], "sogouqa.test.jsonl")],
            vocab_file=VOCAB_PATH,
            limit=32,
            context_key="passage",
        )
        model = BertForQuestionAnswering.from_pretrained(BERT_PATH)
        callback.model = model
        callback.on_epoch_end(epoch=0, logs=None)


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from tokenizers import BertWordPieceTokenizer
from transformers_keras.sequence_classification.dataset import (
    SequenceClassificationDataset,
    SequenceClassificationExample,
)


class DatasetTest(unittest.TestCase):
    """Dataset test."""

    def setUp(self) -> None:
        self.tokenizer = BertWordPieceTokenizer.from_file("testdata/vocab.bert.txt")

    def _read_examples(self, input_files):
        examples = []
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    instance = json.loads(line)
                    encoding = self.tokenizer.encode(instance["sequence"])
                    examples.append(
                        SequenceClassificationExample(
                            tokens=encoding.tokens,
                            input_ids=encoding.ids,
                            segment_ids=encoding.type_ids,
                            attention_mask=encoding.attention_mask,
                            label=int(instance["label"]),
                        )
                    )
        return examples

    def test_sequence_classification_dataset_examples(self):
        examples = self._read_examples(["testdata/sequence_classify.jsonl"])
        dataset = SequenceClassificationDataset.from_examples(
            examples,
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))

        SequenceClassificationDataset.examples_to_tfrecord(examples, ["testdata/sequence_classify.tfrecord"])
        dataset = SequenceClassificationDataset.from_tfrecord_files(
            ["testdata/sequence_classify.tfrecord"] * 4,
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))

    def test_sequence_classification_dataset_jsonl(self):
        dataset = SequenceClassificationDataset.from_jsonl_files(
            "testdata/sequence_classify.jsonl",
            "testdata/vocab.bert.txt",
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))

        SequenceClassificationDataset.jsonl_to_tfrecord(
            "testdata/sequence_classify.jsonl",
            "testdata/vocab.bert.txt",
            "testdata/sequence_classify.tfrecord",
        )
        dataset = SequenceClassificationDataset.from_tfrecord_files(
            ["testdata/sequence_classify.tfrecord"] * 4,
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
from transformers_keras.modeling_bert import BertModel, BertPretrainedModel


class AbstractSimCSE(BertPretrainedModel):

    def __init__(self,
                 vocab_size=21128,
                 max_positions=512,
                 hidden_size=768,
                 type_vocab_size=2,
                 num_layers=6,
                 num_attention_heads=8,
                 intermediate_size=3072,
                 activation='gelu',
                 hidden_dropout_rate=0.2,
                 attention_dropout_rate=0.1,
                 initializer_range=0.02,
                 epsilon=1e-12,
                 **kwargs):
        # build functional model
        input_ids = tf.keras.layers.Input(shape=(None, ), dtype=tf.int32, name='input_ids')
        segment_ids = tf.keras.layers.Input(shape=(None, ), dtype=tf.int32, name='segment_ids')
        attention_mask = tf.keras.layers.Input(shape=(None, ), dtype=tf.int32, name='attention_mask')
        bert_model = BertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            hidden_size=hidden_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name='bert')
        sequence_output, _, _, _ = bert_model(input_ids, segment_ids, attention_mask)
        embedding = tf.keras.layers.Dense(hidden_size, name='embedding')(sequence_output[:, 0, :])
        super().__init__(inputs=[input_ids, segment_ids, attention_mask], outputs=[embedding])
        self.bert_model = bert_model

    def forward(self, x, y, training=False):
        raise NotImplementedError()

    def train_step(self, data):
        x = data
        batch_size = tf.shape(x['input_ids'])[0]
        y_true = tf.range(0, batch_size)
        with tf.GradientTape() as tape:
            y_pred, loss = self.forward(x, y_true, training=True)

        # Compute gradients
        trainable_vars = self.trainable_variables
        # print_variables(trainable_vars)
        gradients = tape.gradient(loss, trainable_vars)
        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y_true, y_pred)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({'loss': loss})
        return results

    def test_step(self, data):
        x = data
        batch_size = tf.shape(x['input_ids'])[0]
        y_true = tf.range(0, batch_size)
        y_pred, loss = self.forward(x, y_true, training=False)
        self.compiled_metrics.update_state(y_true, y_pred)
        results = {m.name: m.result() for m in self.metrics}
        results.update({'loss': loss})
        return results


class UnsupervisedSimCSE(AbstractSimCSE):

    def forward(self, x, y, training=False):
        input_ids, segment_ids, attn_mask = x['input_ids'], x['segment_ids'], x['attention_mask']
        embedding_a = self(inputs=[input_ids, segment_ids, attn_mask], training=training)
        embedding_b = self(inputs=[input_ids, segment_ids, attn_mask], training=training)
        loss, y_pred = self._compute_contrastive_loss(embedding_a, embedding_b, y)
        return y_pred, loss

    def _compute_contrastive_loss(self, embedding_a, embedding_b, labels):
        norm_embedding_a = tf.linalg.normalize(embedding_a, axis=-1)[0]
        norm_embedding_b = tf.linalg.normalize(embedding_b, axis=-1)[0]
        # shape: (batch_size, batch_size)
        cosine = tf.matmul(norm_embedding_a, norm_embedding_b, transpose_b=True)
        # softmax temperature
        cosine = cosine / 0.05
        loss = tf.keras.losses.sparse_categorical_crossentropy(labels, cosine, from_logits=True)
        return loss, cosine


class SupervisedSimCSE(AbstractSimCSE):

    def forward(self, x, y, training=False):
        input_ids, segment_ids, attention_mask = x['input_ids'], x['segment_ids'], x['attention_mask']
        pos_input_ids, pos_segment_ids, pos_attention_mask = x['pos_input_ids'], x['pos_segment_ids'], x['pos_attention_mask']
        embedding = self(inputs=[input_ids, segment_ids, attention_mask], training=training)
        pos_embedding = self(inputs=[pos_input_ids, pos_segment_ids, pos_attention_mask], training=training)
        loss, y_pred = self._compute_contrastive_loss(embedding, pos_embedding, y)
        return y_pred, loss

    def _compute_contrastive_loss(self, embedding, pos_embedding, labels):
        norm_embedding = tf.linalg.normalize(embedding, axis=-1)[0]
        norm_pos_embedding = tf.linalg.normalize(pos_embedding, axis=-1)[0]
        cosine = tf.matmul(norm_embedding, norm_pos_embedding, transpose_b=True)
        # softmax temperature
        cosine = cosine / 0.05
        loss = tf.keras.losses.sparse_categorical_crossentropy(labels, cosine, from_logits=True)
        return loss, cosine


class HardNegativeSimCSE(AbstractSimCSE):

    def forward(self, x, y, training=False):
        input_ids, segment_ids, attention_mask = x['input_ids'], x['segment_ids'], x['attention_mask']
        pos_input_ids, pos_segment_ids, pos_attention_mask = x['pos_input_ids'], x['pos_segment_ids'], x['pos_attention_mask']
        neg_input_ids, neg_segment_ids, neg_attention_mask = x['neg_input_ids'], x['neg_segment_ids'], x['neg_attention_mask']
        embedding = self(inputs=[input_ids, segment_ids, attention_mask], training=training)
        pos_embedding = self(inputs=[pos_input_ids, pos_segment_ids, pos_attention_mask], training=training)
        neg_embedding = self(inputs=[neg_input_ids, neg_segment_ids, neg_attention_mask], training=training)
        loss, y_pred = self._compute_contrastive_loss(embedding, pos_embedding, neg_embedding, y)
        return y_pred, loss

    def _compute_contrastive_loss(self, embedding, pos_embedding, neg_embedding, labels):
        norm_embedding = tf.linalg.normalize(embedding, axis=-1)[0]
        norm_pos_embedding = tf.linalg.normalize(pos_embedding, axis=-1)[0]
        norm_neg_embedding = tf.linalg.normalize(neg_embedding, axis=-1)[0]
        # shape: (batch_size, batch_size)
        pos_sim = tf.matmul(norm_embedding, norm_pos_embedding, transpose_b=True)
        # shape: (batch_size, batch_size)
        neg_sim = tf.matmul(norm_embedding, norm_neg_embedding, transpose_b=True)
        # shape: (batch_size, batch_size * 2)
        cosine = tf.concat([pos_sim, neg_sim], axis=1)
        pos_weight = tf.zeros_like(pos_sim)
        # negative weight = 0.2
        neg_weight = tf.linalg.diag(tf.ones(tf.shape(neg_sim)[0])) * 0.2
        cosine_weight = tf.concat([pos_weight, neg_weight], axis=1)
        # softmax temperature
        cosine = (cosine + cosine_weight) / 0.05
        loss = tf.keras.losses.sparse_categorical_crossentropy(labels, cosine, from_logits=True)
        return loss, cosine



import json
import re
import unittest

from tokenizers import BertWordPieceTokenizer
from transformers_keras.token_classification.dataset import TokenClassificationDataset, TokenClassificationExample
from transformers_keras.token_classification.tokenizer import TokenClassificationLabelTokenizer


class DatasetTest(unittest.TestCase):
    """Dataset tests."""

    def setUp(self) -> None:
        self.tokenizer = BertWordPieceTokenizer.from_file("testdata/vocab.bert.txt")
        self.label_tokenizer = TokenClassificationLabelTokenizer.from_file("testdata/labels.txt")

    def _read_examples(self, input_files, **kwargs):
        examples = []
        for f in input_files:
            features, labels = [], []
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        examples.append(self._compose_example(features, labels, **kwargs))
                        features, labels = [], []
                        continue
                    parts = re.split("\\s+", line)
                    if len(parts) != 2:
                        continue
                    features.append(parts[0])
                    labels.append(parts[1])
            if features and labels:
                examples.append(self._compose_example(features, labels, **kwargs))
        return examples

    def _compose_example(self, features, labels, **kwargs):
        ids = [101] + [self.tokenizer.token_to_id(x) for x in features] + [102]
        label_ids = self.label_tokenizer.labels_to_ids(labels, add_cls=True, add_sep=True)
        return TokenClassificationExample(
            tokens=features,
            input_ids=ids,
            segment_ids=[0] * len(ids),
            attention_mask=[1] * len(ids),
            labels=labels,
            label_ids=label_ids,
        )

    def test_generate_jsonl(self):
        examples = self._read_examples(["testdata/conll.txt"])
        with open("testdata/token_classify.jsonl", mode="wt", encoding="utf-8") as fout:
            for e in examples:
                info = {"features": e.tokens, "labels": e.labels}
                fout.write(json.dumps(info, ensure_ascii=False))
                fout.write("\n")

    def test_dataset(self):
        # test from examples
        examples = self._read_examples(["testdata/conll.txt"])
        dataset = TokenClassificationDataset.from_examples(
            examples,
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))

        # test examples to tfrecord
        TokenClassificationDataset.examples_to_tfrecord(examples, ["testdata/conll.tfrecord"])

        # test conll to tfrecord
        TokenClassificationDataset.conll_to_tfrecord(
            ["testdata/conll.txt"], "testdata/vocab.bert.txt", "testdata/labels.txt", ["testdata/conll.tfrecord"]
        )

        # test jsonl to tfrecord
        TokenClassificationDataset.jsonl_to_tfrecord(
            ["testdata/token_classify.jsonl"],
            "testdata/vocab.bert.txt",
            "testdata/labels.txt",
            ["testdata/token_classify.tfrecord"],
        )

        # test from conll files
        dataset = TokenClassificationDataset.from_conll_files(
            ["testdata/conll.txt"], "testdata/vocab.bert.txt", "testdata/labels.txt", batch_size=4
        )
        print()
        print(next(iter(dataset)))

        # test from jsonl files
        dataset = TokenClassificationDataset.from_jsonl_files(
            ["testdata/token_classify.jsonl"],
            "testdata/vocab.bert.txt",
            "testdata/labels.txt",
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))

        # test from tfrecord
        dataset = TokenClassificationDataset.from_tfrecord_files(
            ["testdata/conll.tfrecord", "testdata/token_classify.tfrecord"],
            batch_size=4,
        )
        print()
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import json
import logging

import tensorflow as tf

from transformers_keras.adapters import parse_pretrained_model_files
from transformers_keras.adapters.albert_adapter import AlbertAdapter

from .modeling_utils import choose_activation


class AlbertEmbedding(tf.keras.layers.Layer):
    """Embeddings for Albert."""

    def __init__(
        self,
        vocab_size=21128,
        max_positions=512,
        type_vocab_size=2,
        embedding_size=128,
        hidden_dropout_rate=0.2,
        initializer_range=0.02,
        epsilon=1e-8,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.vocab_size = vocab_size
        self.max_positions = max_positions
        self.type_vocab_size = type_vocab_size
        self.embedding_size = embedding_size
        self.initializer_range = initializer_range

        self.layernorm = tf.keras.layers.LayerNormalization(epsilon=epsilon, name="LayerNorm")
        self.dropout = tf.keras.layers.Dropout(hidden_dropout_rate)

    def build(self, input_shape):
        self.word_embedding = self.add_weight(
            name="word_embeddings",
            shape=[self.vocab_size, self.embedding_size],
            initializer=tf.keras.initializers.TruncatedNormal(stddev=self.initializer_range),
        )
        self.position_embedding = self.add_weight(
            name="position_embeddings",
            shape=[self.max_positions, self.embedding_size],
            initializer=tf.keras.initializers.TruncatedNormal(stddev=self.initializer_range),
        )
        self.segment_embedding = self.add_weight(
            name="token_type_embeddings",
            shape=[self.type_vocab_size, self.embedding_size],
            initializer=tf.keras.initializers.TruncatedNormal(stddev=self.initializer_range),
        )
        super().build(input_shape)

    def call(self, input_ids, segment_ids=None, position_ids=None, training=None):
        if segment_ids is None:
            segment_ids = tf.zeros_like(input_ids)
        if position_ids is None:
            position_ids = tf.range(0, tf.shape(input_ids)[1], dtype=input_ids.dtype)
            position_ids = tf.expand_dims(position_ids, axis=0)

        position_embeddings = tf.gather(self.position_embedding, position_ids)
        position_embeddings = tf.tile(position_embeddings, multiples=[tf.shape(input_ids)[0], 1, 1])
        token_type_embeddings = tf.gather(self.segment_embedding, segment_ids)
        token_embeddings = tf.gather(self.word_embedding, input_ids)

        embeddings = token_embeddings + token_type_embeddings + position_embeddings
        embeddings = self.layernorm(embeddings)
        embeddings = self.dropout(embeddings, training=training)
        return embeddings


class AlbertMultiHeadAttention(tf.keras.layers.Layer):
    """Multi Head Attention."""

    def __init__(
        self,
        hidden_size=768,
        num_attention_heads=8,
        hidden_dropout_rate=0.0,
        attention_dropout_rate=0.0,
        initializer_range=0.02,
        epsilon=1e-8,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.num_attention_heads = num_attention_heads
        self.hidden_size = hidden_size

        self.query_weight = tf.keras.layers.Dense(self.hidden_size, name="query")
        self.key_weight = tf.keras.layers.Dense(self.hidden_size, name="key")
        self.value_weight = tf.keras.layers.Dense(self.hidden_size, name="value")

        self.dense = tf.keras.layers.Dense(
            hidden_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            name="dense",
        )
        self.layer_norm = tf.keras.layers.LayerNormalization(epsilon=epsilon, name="layer_norm")
        self.attention_dropout = tf.keras.layers.Dropout(attention_dropout_rate)
        self.output_dropout = tf.keras.layers.Dropout(hidden_dropout_rate)

    def _split_heads(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_attention_heads, self.hidden_size // self.num_attention_heads))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def _scaled_dot_product_attention(self, query, key, value, attention_mask, training=None):
        query = tf.cast(query, dtype=self.dtype)
        key = tf.cast(key, dtype=self.dtype)
        value = tf.cast(value, dtype=self.dtype)

        score = tf.matmul(query, key, transpose_b=True)
        dk = tf.cast(tf.shape(query)[-1], self.dtype)
        score = score / tf.math.sqrt(dk)
        if attention_mask is not None:
            attention_mask = tf.cast(attention_mask, dtype=self.dtype)
            score += tf.cast((1.0 - attention_mask) * -10000.0, self.dtype)
        attn_weights = tf.nn.softmax(score, axis=-1)
        attn_weights = self.attention_dropout(attn_weights, training=training)
        context = tf.matmul(attn_weights, value)
        return context, attn_weights

    def call(self, query, key, value, attention_mask, training=None):
        origin_input = query  # query == key == value

        batch_size = tf.shape(query)[0]
        query = self._split_heads(self.query_weight(query), batch_size)
        key = self._split_heads(self.key_weight(key), batch_size)
        value = self._split_heads(self.value_weight(value), batch_size)

        context, attn_weights = self._scaled_dot_product_attention(query, key, value, attention_mask, training=training)
        context = tf.transpose(context, perm=[0, 2, 1, 3])
        context = tf.reshape(context, [batch_size, -1, self.hidden_size])
        output = self.dense(context)
        output = self.output_dropout(output, training=training)
        output = self.layer_norm(output + origin_input)
        return output, attn_weights


class AlbertEncoderLayer(tf.keras.layers.Layer):
    """Encoder layer."""

    def __init__(
        self,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.0,
        attention_dropout_rate=0.0,
        epsilon=1e-8,
        initializer_range=0.02,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.attention = AlbertMultiHeadAttention(
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="attention",
        )

        self.ffn = tf.keras.layers.Dense(
            intermediate_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            name="ffn",
        )
        self.activation = choose_activation(activation)
        self.ffn_output = tf.keras.layers.Dense(
            hidden_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            name="ffn_output",
        )
        self.layer_norm = tf.keras.layers.LayerNormalization(epsilon=epsilon, name="layer_norm")
        self.dropout = tf.keras.layers.Dropout(hidden_dropout_rate)

    def call(self, hidden_states, attention_mask, training=None):
        attn_output, attn_weights = self.attention(
            hidden_states, hidden_states, hidden_states, attention_mask, training=training
        )
        outputs = self.ffn(attn_output)
        outputs = self.activation(outputs)
        outputs = self.ffn_output(outputs)
        outputs = self.dropout(outputs)
        outputs = self.layer_norm(outputs + attn_output)
        return outputs, attn_weights


class AlbertEncoderGroup(tf.keras.layers.Layer):
    """Encoder group."""

    def __init__(
        self,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.encoder_layers = [
            AlbertEncoderLayer(
                hidden_size=hidden_size,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                activation=activation,
                hidden_dropout_rate=hidden_dropout_rate,
                attention_dropout_rate=attention_dropout_rate,
                epsilon=epsilon,
                initializer_range=initializer_range,
                name="layer_{}".format(i),
            )
            for i in range(num_layers_each_group)
        ]

    def call(self, hidden_states, attention_mask, training=None):
        group_hidden_states, group_attn_weights = [], []
        for idx, encoder in enumerate(self.encoder_layers):
            hidden_states, attn_weights = encoder(hidden_states, attention_mask)
            group_hidden_states.append(hidden_states)
            group_attn_weights.append(attn_weights)

        return hidden_states, group_hidden_states, group_attn_weights


class AlbertEncoder(tf.keras.layers.Layer):
    """Encoder, stack of encoder groups."""

    def __init__(
        self,
        num_layers=12,
        num_groups=1,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        **kwargs
    ):
        super(AlbertEncoder, self).__init__(**kwargs)

        self.embedding_mapping = tf.keras.layers.Dense(
            hidden_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            name="embedding_mapping",
        )
        self.groups = [
            AlbertEncoderGroup(
                num_layers_each_group=num_layers_each_group,
                hidden_size=hidden_size,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                activation=activation,
                hidden_dropout_rate=hidden_dropout_rate,
                attention_dropout_rate=attention_dropout_rate,
                epsilon=epsilon,
                initializer_range=initializer_range,
                name="group_{}".format(i),
            )
            for i in range(num_groups)
        ]

        self.num_layers = num_layers
        self.num_groups = num_groups

    def call(self, hidden_states, attention_mask, training=None):
        hidden_states = self.embedding_mapping(hidden_states)

        all_hidden_states, all_attention_weights = [], []
        for i in range(self.num_layers):
            layers_per_group = self.num_layers // self.num_groups
            group_index = i // layers_per_group
            hidden_states, group_hidden_states, group_attn_weights = self.groups[group_index](
                hidden_states, attention_mask
            )
            all_hidden_states.extend(group_hidden_states)
            all_attention_weights.extend(group_attn_weights)

        # stack all_hidden_states to shape:
        # [batch_size, num_layers, num_attention_heads, hidden_size]
        all_hidden_states = tf.stack(all_hidden_states, axis=0)
        all_hidden_states = tf.transpose(all_hidden_states, perm=[1, 0, 2, 3])
        # stack all_attention_scores to shape:
        # [batch_size, num_layers, num_attention_heads, seqeucen_length, sequence_length]
        all_attention_weights = tf.stack(all_attention_weights, axis=0)
        all_attention_weights = tf.transpose(all_attention_weights, perm=[1, 0, 2, 3, 4])
        return hidden_states, all_hidden_states, all_attention_weights


class AlbertPooler(tf.keras.layers.Layer):
    """Pooler."""

    def __init__(self, hidden_size, initializer_range=0.02, **kwargs):
        super().__init__(**kwargs)
        self.dense = tf.keras.layers.Dense(
            hidden_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            activation="tanh",
            name="dense",
        )

    def call(self, sequence_output):
        return self.dense(sequence_output[:, 0])


class AlbertModel(tf.keras.layers.Layer):
    """Albert model."""

    def __init__(
        self,
        vocab_size=21128,
        max_positions=512,
        embedding_size=128,
        type_vocab_size=2,
        num_layers=12,
        num_groups=1,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.embedding = AlbertEmbedding(
            vocab_size=vocab_size,
            max_positions=max_positions,
            embedding_size=embedding_size,
            type_vocab_size=type_vocab_size,
            hidden_dropout_rate=hidden_dropout_rate,
            epsilon=epsilon,
            initializer_range=initializer_range,
            name="embeddings",
        )
        self.encoder = AlbertEncoder(
            num_layers=num_layers,
            num_groups=num_groups,
            num_layers_each_group=num_layers_each_group,
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            epsilon=epsilon,
            initializer_range=initializer_range,
            name="encoder",
        )
        self.pooler = AlbertPooler(hidden_size=hidden_size, initializer_range=initializer_range, name="pooler")

    def call(self, input_ids, segment_ids, attention_mask, training=None):
        attention_mask = attention_mask[:, tf.newaxis, tf.newaxis, :]
        embed = self.embedding(input_ids, segment_ids)
        sequence_outputs, all_hidden_states, all_attn_weights = self.encoder(embed, attention_mask)
        # take [CLS]
        pooled_output = self.pooler(sequence_outputs)
        return sequence_outputs, pooled_output, all_hidden_states, all_attn_weights


class AlbertPretrainedModel(tf.keras.Model):
    """Base model for pretrained albert models. All subclass models can load pretrained model directly."""

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_dir,
        override_params=None,
        adapter=None,
        use_functional_api=True,
        check_weights=True,
        verbose=True,
        **kwargs
    ):
        """Load pretrained albert weights.

        Args:
            pretrained_model_dir: A file path to pretrained model dir, including ckpt and config files.
            override_params: Add extra model params or override pretrained model's params to constructor
            use_functional_api: Python boolean, default is True.
                When subclassing BertPretrainedModel, you may construct model using Keras' functional API,
                then you should set use_functional_api=True, otherwise, you should set use_functional_api=False.
            adapter: An adpater to adapte pretrained weights to this new created model. Default is `AlbertAdapter`.
                In most case, you do not need to specify a `adapter`.
            check_weights: Python boolean. If true, check model weights' value after loading weights from ckpt
            verbose: Python boolean.If True, logging more detailed informations when loadding pretrained weights
        """

        config_file, ckpt, _ = parse_pretrained_model_files(pretrained_model_dir)
        if not adapter:
            adapter = AlbertAdapter(
                skip_token_embedding=kwargs.pop("skip_token_embedding", False),
                skip_position_embedding=kwargs.pop("skip_position_embedding", False),
                skip_segment_embedding=kwargs.pop("skip_segment_embedding", False),
                skip_embedding_layernorm=kwargs.pop("skip_embedding_layernorm", False),
                skip_pooler=kwargs.pop("skip_pooler", False),
            )
        model_config = adapter.adapte_config(config_file, **kwargs)
        if override_params:
            model_config.update(override_params)
        logging.info("Load model config: \n%s", json.dumps(model_config, indent=4))
        model = cls(**model_config, **kwargs)
        albert_model = getattr(model, "albert_model", None)
        assert albert_model is not None, "AlbertPretrainedModel must have an attribute named albert_model!"
        inputs = model.dummy_inputs()
        model(inputs=list(inputs), training=False)
        adapter.adapte_weights(
            albert=model.albert_model,
            config=model_config,
            ckpt=ckpt,
            prefix="" if use_functional_api else model.name,
            check_weights=check_weights,
            verbose=verbose,
            **kwargs
        )
        return model

    @classmethod
    def from_config_file(cls, config_file, override_params=None, adapter=None, **kwargs):
        if not adapter:
            adapter = AlbertAdapter(
                skip_token_embedding=kwargs.pop("skip_token_embedding", False),
                skip_position_embedding=kwargs.pop("skip_position_embedding", False),
                skip_segment_embedding=kwargs.pop("skip_segment_embedding", False),
                skip_embedding_layernorm=kwargs.pop("skip_embedding_layernorm", False),
            )
        model_config = adapter.adapte_config(config_file, **kwargs)
        if override_params:
            model_config.update(override_params)
        logging.info("Load model config: \n%s", json.dumps(model_config, indent=4))
        model = cls(**model_config, **kwargs)
        albert_model = getattr(model, "albert_model", None)
        assert albert_model is not None, "AlbertPretrainedModel must have an attribute named albert_model!"
        inputs = model.dummy_inputs()
        model(inputs=list(inputs), training=False)
        return model

    def dummy_inputs(self):
        input_ids = tf.constant([0] * 128, dtype=tf.int64, shape=(1, 128))
        segment_ids = tf.constant([0] * 128, dtype=tf.int64, shape=(1, 128))
        attn_mask = tf.constant([1] * 128, dtype=tf.int64, shape=(1, 128))
        return input_ids, segment_ids, attn_mask


class Albert(AlbertPretrainedModel):
    """Albert model, can load pretrained models directly."""

    def __init__(
        self,
        vocab_size=21128,
        max_positions=512,
        embedding_size=128,
        type_vocab_size=2,
        num_layers=12,
        num_groups=1,
        num_layers_each_group=1,
        hidden_size=768,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        epsilon=1e-12,
        initializer_range=0.02,
        return_states=False,
        return_attention_weights=False,
        **kwargs
    ):
        # build functional model
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        albert_model = AlbertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            embedding_size=embedding_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_groups=num_groups,
            num_layers_each_group=num_layers_each_group,
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            epsilon=epsilon,
            initializer_range=initializer_range,
            name="albert",
        )
        sequence_output, pooled_output, hidden_states, attention_weights = albert_model(
            input_ids, segment_ids, attention_mask
        )
        outputs = [
            tf.keras.layers.Lambda(lambda x: x, name="sequence_output")(sequence_output),
            tf.keras.layers.Lambda(lambda x: x, name="pooled_output")(pooled_output),
        ]
        if return_states:
            outputs += [tf.keras.layers.Lambda(lambda x: x, name="hidden_states")(hidden_states)]
        if return_attention_weights:
            outputs += [tf.keras.layers.Lambda(lambda x: x, name="attention_weights")(attention_weights)]

        super().__init__(inputs=[input_ids, segment_ids, attention_mask], outputs=outputs, **kwargs)

        self.albert_model = albert_model
        self.vocab_size = vocab_size
        self.type_vocab_size = type_vocab_size
        self.max_positions = max_positions
        self.embedding_size = embedding_size
        self.num_layers = num_layers
        self.num_groups = num_groups
        self.num_layers_each_group = num_layers_each_group
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.intermediate_size = intermediate_size
        self.hidden_dropout_rate = hidden_dropout_rate
        self.attention_dropout_rate = attention_dropout_rate
        self.initializer_range = initializer_range
        self.return_states = return_states
        self.return_attention_weights = return_attention_weights

    def get_config(self):
        config = {
            "vocab_size": self.vocab_size,
            "type_vocab_size": self.type_vocab_size,
            "max_positions": self.max_positions,
            "num_layers": self.num_layers,
            "hidden_size": self.hidden_size,
            "num_attention_heads": self.num_attention_heads,
            "intermediate_size": self.intermediate_size,
            "hidden_dropout_rate": self.hidden_dropout_rate,
            "attention_dropout_rate": self.attention_dropout_rate,
            "initializer_range": self.initializer_range,
            "return_states": self.return_states,
            "return_attention_weights": self.return_attention_weights,
        }
        return config



class TokenClassificationLabelTokenizer:
    """Label tokenizer."""

    def __init__(self, label2id, o_token="O", **kwargs) -> None:
        self.label2id = label2id
        self.id2label = {v: k for k, v in label2id.items()}
        self.o_token = o_token
        self.o_id = self.label2id[o_token]

    def label_to_id(self, token, **kwargs):
        return self.label2id.get(token, self.o_id)

    def id_to_label(self, id, **kwargs):
        return self.id2label.get(id, self.o_token)

    def labels_to_ids(self, tokens, add_cls=False, add_sep=False, **kwargs):
        ids = [self.label2id.get(token, self.o_id) for token in tokens]
        if add_cls:
            ids = [self.o_id] + ids
        if add_sep:
            ids = ids + [self.o_id]
        return ids

    def ids_to_labels(self, ids, del_cls=True, del_sep=True, **kwargs):
        tokens = [self.id2label.get(_id, self.o_token) for _id in ids]
        if tokens and del_cls and tokens[0] == self.o_token:
            tokens = tokens[1:]
        if tokens and del_sep and tokens[-1] == self.o_token:
            tokens = tokens[:-1]
        return tokens

    @classmethod
    def from_file(cls, vocab_file, o_token="O", **kwargs):
        label2id = cls._read_label_vocab(vocab_file)
        return cls(label2id=label2id, o_token=o_token, **kwargs)

    @classmethod
    def _read_label_vocab(cls, vocab_file):
        m = {}
        with open(vocab_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                k = line.strip()
                m[k] = len(m)
        return m



import abc
import json
import logging
import os

import tensorflow as tf


def zip_weights(model, ckpt, variables_mapping, **kwargs):
    weights, values = [], []
    for w in model.trainable_weights:
        var = variables_mapping.get(w.name, None)
        if var is None:
            logging.warning('Model weight: %s not collected in weights mapping.', w.name)
            continue
        v = tf.train.load_variable(ckpt, var)
        if w.name == 'bert/nsp/dense/kernel:0':
            v = v.T
        weights.append(w)
        values.append(v)
        if kwargs.get('verbose', True):
            logging.info('Load weight: {:60s} <-- {}'.format(w.name, variables_mapping[w.name]))

    mapped_values = zip(weights, values)
    return mapped_values


def parse_pretrained_model_files(pretrained_model_dir):
    config_file, ckpt, vocab = None, None, None
    pretrained_model_dir = os.path.abspath(pretrained_model_dir)
    if not os.path.exists(pretrained_model_dir):
        logging.info('pretrain model dir: {} is not exists.'.format(pretrained_model_dir))
        return config_file, ckpt, vocab
    for f in os.listdir(pretrained_model_dir):
        if str(f).endswith('config.json'):
            config_file = os.path.join(pretrained_model_dir, f)
        if 'vocab' in str(f):
            vocab = os.path.join(pretrained_model_dir, f)
        if 'ckpt' in str(f):
            n = '.'.join(str(f).split('.')[:-1])
            ckpt = os.path.join(pretrained_model_dir, n)
    return config_file, ckpt, vocab


class AbstractAdapter(abc.ABC):

    def __init__(self,
                 skip_token_embedding=False,
                 skip_position_embedding=False,
                 skip_segment_embedding=False,
                 skip_embedding_layernorm=False,
                 skip_pooler=False,
                 **kwargs):
        super().__init__()
        self.skip_token_embedding = skip_token_embedding
        self.skip_position_embedding = skip_position_embedding
        self.skip_segment_embedding = skip_segment_embedding
        self.skip_embedding_layernorm = skip_embedding_layernorm
        self.skip_pooler = skip_pooler

        logging.info('Adapter skipping config: %s', json.dumps({
            'skip_token_embedding': self.skip_token_embedding,
            'skip_position_embedding': self.skip_position_embedding,
            'skip_segment_embedding': self.skip_segment_embedding,
            'skip_embedding_layernorm': self.skip_embedding_layernorm,
            'skip_pooler': self.skip_pooler
        }))

    @abc.abstractmethod
    def adapte_config(self, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def adapte_weights(self, *args, **kwargs):
        raise NotImplementedError()



import os
import unittest

import tensorflow as tf
from transformers_keras.sentence_embedding.simcse_models import (
    HardNegativeSimCSE, SupervisedSimCSE, UnsupervisedSimCSE)


class SimCSEModelsTest(unittest.TestCase):

    def test_build_simcse_models(self):
        m1 = UnsupervisedSimCSE()
        m1.summary()

        m2 = SupervisedSimCSE()
        m2.summary()

        m3 = HardNegativeSimCSE()
        m3.summary()

    def test_simcse_models_load_pretrained(self):

        def _run_model(model, name):
            model.summary()
            p = os.path.join('models/simcse-{}/1'.format(name))
            model.save(p, include_optimizer=False)

            loaded = tf.saved_model.load(p)
            model = loaded.signatures['serving_default']
            print(model.structured_input_signature)
            print(model.structured_outputs)

        bert_path = os.path.join(os.environ['CHINESE_BERT_PATH'], 'chinese_roberta_wwm_ext_L-12_H-768_A-12')
        m1 = UnsupervisedSimCSE.from_pretrained(bert_path)
        _run_model(m1, 'unsup')

        m2 = SupervisedSimCSE.from_pretrained(bert_path)
        _run_model(m2, 'sup')

        m3 = HardNegativeSimCSE.from_pretrained(bert_path)
        _run_model(m3, 'hardneg')


if __name__ == "__main__":
    unittest.main()
